import datetime
import errno
import inspect
import json
import os
import platform
import socket

from django.db import connections
from django.test.client import ClientHandler
from django.test.testcases import LiveServerTestCase, LiveServerThread
from google.appengine.api import apiproxy_stub_map, datastore_types
from google.appengine.ext import deferred, ndb, testbed
from rest_framework import status
from rest_framework.test import APITestCase, APITransactionTestCase

from freezegun import freeze_time as fg_freeze_time
from freezegun.api import FakeDatetime
from jsonschema import Draft4Validator
from mock import patch

_is_cpython = (
    hasattr(platform, 'python_implementation') and
    platform.python_implementation().lower() == "cpython"
)


# Need any async tasklets to run before request is complete since
# Django tests don't run through real wsgi handler we fake it here.
old_get_response = ClientHandler.get_response
@ndb.toplevel
def get_response(self, *args, **kwargs):
    return old_get_response(self, *args, **kwargs)
ClientHandler.get_response = get_response


class HTTPStatusTestCaseMixin(object):

    def assertHTTP200(self, response):
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def assertHTTP201(self, response):
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def assertHTTP204(self, response):
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def assertHTTP400(self, response):
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def assertHTTP401(self, response):
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def assertHTTP403(self, response):
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def assertHTTP404(self, response):
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class _freeze_time(object):
    def __init__(self, *args, **kwargs):
        self._gae_patch = patch('google.appengine.ext.db.DateTimeProperty.data_type', new=FakeDatetime)
        self._freeze = fg_freeze_time(*args, **kwargs)
        datastore_types._VALIDATE_PROPERTY_VALUES[FakeDatetime] = datastore_types.ValidatePropertyNothing
        datastore_types._PACK_PROPERTY_VALUES[FakeDatetime] = datastore_types.PackDatetime
        datastore_types._PROPERTY_MEANINGS[FakeDatetime] = datastore_types.entity_pb.Property.GD_WHEN

    def __call__(self, func):
        if inspect.isclass(func):
            return self._freeze.decorate_class(func)
        return self._freeze.decorate_callable(func)

    def start(self):
        self._freeze.start()
        self._gae_patch.start()

    def stop(self):
        self._freeze.stop()
        self._gae_patch.stop()

    def __enter__(self):
        return self.start()

    def __exit__(self, *args):
        self.stop()


def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False):
    # Python3 doesn't have basestring, but it does have str.
    try:
        string_type = basestring
    except NameError:
        string_type = str

    if not isinstance(time_to_freeze, (string_type, datetime.date)):
        raise TypeError(('freeze_time() expected None, a string, date instance, or '
                         'datetime instance, but got type {0}.').format(type(time_to_freeze)))
    if tick and not _is_cpython:
        raise SystemError('Calling freeze_time with tick=True is only compatible with CPython')

    if ignore is None:
        ignore = []
    ignore.append('six.moves')
    ignore.append('django.utils.six.moves')
    return _freeze_time(time_to_freeze, tz_offset, ignore, tick)


class ClientHelperMixin(object):

    def validate(self, data, schema):
        v = Draft4Validator(schema)
        if not v.is_valid(data):
            for error in sorted(v.iter_errors(data), key=str):
                if error.context:
                    self.fail("\n".join('{} {}'.format(cerror.absolute_path, cerror.message) for cerror in error.context))
                else:
                    self.fail(error.message)

    def validate_retrieve(self, response, schema):
        self.validate(response.data, schema)

    def validate_list(self, response, schema, pages=True):
        if pages:
            self.validate(response.data['results'][0], schema['items']['oneOf'][0])
            self.validate(response.data['results'], schema)
        else:
            self.validate(response.data[0], schema['items']['oneOf'][0])
            self.validate(response.data, schema)

    def get(self, path, data=None, follow=False, secure=False, **extra):
        print '***' * 15
        print 'GET {}'.format(path)
        print 'Authenticated' if '_auth_user_id' in self.client.session else 'Not Authenticated'
        response = self.client.get(path=path, data=data, follow=follow, secure=secure, **extra)
        print 'Response status code: {}'.format(response.status_code)
        if hasattr(response, 'data'):
            print 'Response body:'
            print json.dumps(response.data)
        return response

    def post(self, path, data=None, content_type=None, follow=False, secure=False, **extra):
        print '***' * 15
        print 'POST {}'.format(path)
        print 'Authenticated' if '_auth_user_id' in self.client.session else 'Not Authenticated'
        print 'Request body:'
        print json.dumps(data)
        response = self.client.post(path=path, data=data, content_type=content_type, follow=follow, secure=secure,
                                    **extra)
        print 'Response status code: {}'.format(response.status_code)
        print 'Response body:'
        if response.status_code != 204:
            print json.dumps(response.data)
        return response

    def patch(self, path, data=None, content_type=None, follow=False, secure=False, **extra):
        print '***' * 15
        print 'PATCH {}'.format(path)
        print 'Authenticated' if '_auth_user_id' in self.client.session else 'Not Authenticated'
        print 'Request body:'
        print json.dumps(data)
        response = self.client.patch(path=path, data=data, content_type=content_type, follow=follow, secure=secure,
                                     **extra)
        print 'Response status code: {}'.format(response.status_code)
        print 'Response body:'
        print json.dumps(response.data)
        return response

    def put(self, path, data=None, content_type=None, follow=False, secure=False, **extra):
        print '***' * 15
        print 'PUT {}'.format(path)
        print 'Authenticated' if '_auth_user_id' in self.client.session else 'Not Authenticated'
        print 'Request body:'
        print json.dumps(data)
        response = self.client.put(path=path, data=data, content_type=content_type, follow=follow, secure=secure,
                                   **extra)
        print 'Response status code: {}'.format(response.status_code)
        print 'Response body:'
        print json.dumps(response.data)
        return response

    def delete(self, path, data=None, content_type=None, follow=False, secure=False, **extra):
        print '***' * 15
        print 'DELETE {}'.format(path)
        print 'Authenticated' if '_auth_user_id' in self.client.session else 'Not Authenticated'
        print 'Request body:'
        print json.dumps(data)
        response = self.client.delete(path=path, data=data, content_type=content_type, follow=follow, secure=secure,
                                      **extra)
        print 'Response status code: {}'.format(response.status_code)
        print 'Response body:'
        print json.dumps(response.data)
        return response


class GAEHelperMixin(object):

    def setUp(self):

        self.testbed = testbed.Testbed()
        self.testbed.activate()
        path = os.path.normpath(os.path.join(os.path.dirname(__file__), '../../'))
        self.testbed.init_taskqueue_stub(root_path=path)
        self.testbed.init_memcache_stub()
        self.testbed.init_datastore_v3_stub()
        self.taskqueue_stub = self.testbed.get_stub(testbed.TASKQUEUE_SERVICE_NAME)

        self.apiproxy = self.testbed._test_stub_map

        super(GAEHelperMixin, self).setUp()

    @ndb.toplevel
    def sync_tasks(self):
        """
        Because we use ndb.tasklet to submit to tasks to queues there can be one's that haven't submitted yet
        call this method to make sure they get submitted before moving on. We already wrap this in common places
        like using self.post, self.get so you won't need to use it there.
        :return:
        """
        pass

    def delete_task(self, task):
        self.taskqueue_stub.DeleteTask(task.headers['X-AppEngine-QueueName'], task.name)

    def delete_tasks(self, tasks):
        for task in tasks:
            self.delete_task(task)

    def flush_queues(self):
        self.sync_tasks()
        queues = self.taskqueue_stub.GetQueues()
        for queue in queues:
            self.taskqueue_stub.FlushQueue(queue['name'])

    @ndb.toplevel
    def run_taskqueue(self, url):
        tasks = self.taskqueue_stub.get_filtered_tasks(url)
        for task in tasks:
            deferred.run(task.payload)
            self.delete_task(task)

    def run_taskqueues(self, urls):
        for url in urls:
            self.run_taskqueue(url)


class TestCase(HTTPStatusTestCaseMixin, ClientHelperMixin, GAEHelperMixin, APITestCase):
    pass


class TransactionTestCase(HTTPStatusTestCaseMixin, ClientHelperMixin, GAEHelperMixin, APITransactionTestCase):
    pass


class LiveServerThread(LiveServerThread):
    """
    Thread for running a live http server while the tests are running.
    """
    def __init__(self, host, possible_ports, static_handler, connections_override=None):
        from google.appengine.api import apiproxy_stub_map
        self.apiproxy = apiproxy_stub_map.apiproxy

        super(LiveServerThread, self).__init__(host, possible_ports, static_handler, connections_override)

    def run(self):
        """
        Sets up the live server and databases, and then loops over handling
        http requests.
        """
        if self.connections_override:
            # Override this thread's database connections with the ones
            # provided by the main thread.
            for alias, conn in self.connections_override.items():
                connections[alias] = conn

        apiproxy_stub_map.apiproxy = self.apiproxy

        try:
            # This ensures that all ndb tasklets are ran before request is finished
            from django.core.wsgi import get_wsgi_application
            from google.appengine.ext import ndb
            from djangae.wsgi import DjangaeApplication
            handler = ndb.toplevel(DjangaeApplication(get_wsgi_application()))

            # Go through the list of possible ports, hoping that we can find
            # one that is free to use for the WSGI server.
            for index, port in enumerate(self.possible_ports):
                try:
                    self.httpd = self._create_server(port)
                except socket.error as e:
                    if (index + 1 < len(self.possible_ports) and
                                e.errno == errno.EADDRINUSE):
                        # This port is already in use, so we go on and try with
                        # the next one in the list.
                        continue
                    else:
                        # Either none of the given ports are free or the error
                        # is something else than "Address already in use". So
                        # we let that error bubble up to the main thread.
                        raise
                else:
                    # A free port was found.
                    self.port = port
                    break

            self.httpd.set_app(handler)
            self.is_ready.set()
            self.httpd.serve_forever()
        except Exception as e:
            self.error = e
            self.is_ready.set()


class LiveServerTestCase(HTTPStatusTestCaseMixin, ClientHelperMixin, LiveServerTestCase):

    @classmethod
    def setUpClass(cls):

        cls.testbed = testbed.Testbed()
        cls.testbed.activate()
        path = os.path.normpath(os.path.join(os.path.dirname(__file__), '../../'))
        cls.testbed.init_taskqueue_stub(root_path=path)
        cls.testbed.init_memcache_stub()
        cls.testbed.init_datastore_v3_stub()
        cls.taskqueue_stub = cls.testbed.get_stub(testbed.TASKQUEUE_SERVICE_NAME)

        # Make the apiproxy available for other uses
        cls.apiproxy = cls.testbed._test_stub_map

        super(LiveServerTestCase, cls).setUpClass()

    @classmethod
    def _create_server_thread(cls, host, possible_ports, connections_override):
        return LiveServerThread(
            host,
            possible_ports,
            cls.static_handler,
            connections_override=connections_override
        )

    @ndb.toplevel
    def sync_tasks(self):
        """
        Because we use ndb.tasklet to submit to tasks to queues there can be one's that haven't submitted yet
        call this method to make sure they get submitted before moving on. We already wrap this in common places
        like using self.post, self.get so you won't need to use it there.
        :return:
        """
        pass

    def delete_task(self, task):
        self.taskqueue_stub.DeleteTask(task.headers['X-AppEngine-QueueName'], task.name)

    def delete_tasks(self, tasks):
        for task in tasks:
            self.delete_task(task)

    def flush_queues(self):
        self.sync_tasks()
        queues = self.taskqueue_stub.GetQueues()
        for queue in queues:
            self.taskqueue_stub.FlushQueue(queue['name'])

    @ndb.toplevel
    def run_taskqueue(self, url):
        tasks = self.taskqueue_stub.get_filtered_tasks(url)
        for task in tasks:
            deferred.run(task.payload)
            self.delete_task(task)

    def run_taskqueues(self, urls):
        for url in urls:
            self.run_taskqueue(url)



import re

from django_enumfield import enum
from rest_framework import serializers
from rest_framework.utils.field_mapping import get_field_kwargs, ClassLookupDict
import base32_crockford


class Base32CrockfordField(serializers.CharField):

    def to_internal_value(self, data):
        data = super(Base32CrockfordField, self).to_internal_value(data)
        try:
            return base32_crockford.normalize(data)
        except ValueError:
            raise serializers.ValidationError('Invalid {}'.format(self.source))


class USPhoneNumberField(serializers.CharField):

    def to_internal_value(self, data):
        data = re.sub('[^0-9]', '', data)
        data_length = len(data)

        if 10 > data_length > 11:
            raise serializers.ValidationError('The {} should be 10 to 11 numbers in length.'.format(self.source))

        # Prepend the country code
        if data_length == 10:
            data = '+1{}'.format(data)
        else:
            data = '+{}'.format(data)

        return data


class DateTimeSerializerMixin(serializers.Serializer):

    created_at = serializers.DateTimeField(read_only=True)
    updated_at = serializers.DateTimeField(read_only=True)


class PrimaryKeyMixin(serializers.ModelSerializer):
    id = serializers.CharField(read_only=True)


class WritableIDSerializer(serializers.Serializer):
    id = serializers.CharField()


class ObjectWithIDSerializer(serializers.Serializer):
    id = serializers.CharField()


class EnumField(serializers.ChoiceField):
    """ A field that takes a field's value as the key and returns
    the associated value for serialization """
    def __init__(self, choices, *args, **kwargs):
        self.representations = {}
        self.internal_values = {}

        for k, v in choices:
            if v in self.internal_values:
                raise ValueError(
                    'The field is not deserializable with the given choices.'
                    ' Please ensure that choices map 1:1 with values'
                )
            self.internal_values[v.name] = k
            self.representations[k] = v.name

        super(EnumField, self).__init__(self.internal_values, **kwargs)

    def to_internal_value(self, data):
        res = super(EnumField, self).to_internal_value(data)
        return self.internal_values.get(data) if res else 3

    def to_representation(self, value):
        return self.representations.get(value)


class FileProxyURLField(serializers.Field):
    """
    Allows you to proxy through services like Imgix back to where your original file storage location is.

    Mostly just wraps up replacing the URL returned with the proper domain (proxy's domain), and stripping out any
    path prefix if necessary incase your proxy is setup to dive deeper by default.
    """

    def __init__(self, proxy_base_url, strip_prefix=None, *args, **kwargs):
        self.proxy_base_url = proxy_base_url
        self.strip_prefix = strip_prefix
        super(FileProxyURLField, self).__init__(*args, **kwargs)

    def to_representation(self, file):
        # Handle local tests/fixtures that store the URL to an image on randomuser.me
        # Get storage and remove base_url to test
        if file:
            base_filepath = file.url.replace(file.storage.base_url, '')
        else:
            return None

        if base_filepath.startswith('http'):
            url = base_filepath
        else:
            if self.proxy_base_url:
                if self.strip_prefix:
                    base_filepath = base_filepath.replace('{}/'.format(self.strip_prefix), '')
                url = '{}{}'.format(self.proxy_base_url, base_filepath)
            else:
                url = file.url
                request = self.context.get('request', None)
                if request is not None:
                    url = request.build_absolute_uri(url)

        return url


class Serializer(PrimaryKeyMixin, serializers.Serializer):

    def __init__(self, *args, **kwargs):
        nested_write = kwargs.pop('nested_write', False)
        if nested_write:
            for k, field in self.fields.items():
                if field.field_name == 'id':
                    field.read_only = False
                else:
                    field.read_only = True
        super(Serializer, self).__init__(*args, **kwargs)

    def build_standard_field(self, field_name, model_field):
        field_mapping = ClassLookupDict(self.serializer_field_mapping)

        field_class = field_mapping[model_field]
        field_kwargs = get_field_kwargs(field_name, model_field)

        if not issubclass(field_class, serializers.ModelField):
            # `model_field` is only valid for the fallback case of
            # `ModelField`, which is used when no other typed field
            # matched to the model field.
            field_kwargs.pop('model_field', None)

        if field_class == EnumField:
            return field_class, field_kwargs
        else:
            return super(Serializer, self).build_standard_field(field_name, model_field)


class ModelSerializer(PrimaryKeyMixin, serializers.ModelSerializer):
    serializer_field_mapping = serializers.ModelSerializer.serializer_field_mapping
    serializer_field_mapping[enum.EnumField] = EnumField

    def __init__(self, *args, **kwargs):
        nested_write = kwargs.pop('nested_write', False)
        if nested_write:
            for k, field in self.fields.items():
                if field.field_name == 'id':
                    field.read_only = False
                else:
                    field.read_only = True
        super(ModelSerializer, self).__init__(*args, **kwargs)

    def get_default_field_names(self, declared_fields, model_info):
        """
        Return the default list of field names that will be used if the
        `Meta.fields` option is not specified.
        """
        return (
            [model_info.pk.name] +
            [self.url_field_name] +
            list(declared_fields.keys()) +
            list(model_info.fields.keys()) +
            list(model_info.forward_relations.keys())
        )

    def build_standard_field(self, field_name, model_field):
        field_mapping = ClassLookupDict(self.serializer_field_mapping)

        field_class = field_mapping[model_field]
        field_kwargs = get_field_kwargs(field_name, model_field)

        if not issubclass(field_class, serializers.ModelField):
            # `model_field` is only valid for the fallback case of
            # `ModelField`, which is used when no other typed field
            # matched to the model field.
            field_kwargs.pop('model_field', None)

        if field_class == EnumField:
            return field_class, field_kwargs
        else:
            return super(ModelSerializer, self).build_standard_field(field_name, model_field)


class HyperlinkedModelSerializer(serializers.HyperlinkedModelSerializer):
    serializer_field_mapping = serializers.HyperlinkedModelSerializer.serializer_field_mapping
    serializer_field_mapping[enum.EnumField] = EnumField

    def build_standard_field(self, field_name, model_field):
        field_mapping = ClassLookupDict(self.serializer_field_mapping)

        field_class = field_mapping[model_field]
        field_kwargs = get_field_kwargs(field_name, model_field)

        if not issubclass(field_class, serializers.ModelField):
            # `model_field` is only valid for the fallback case of
            # `ModelField`, which is used when no other typed field
            # matched to the model field.
            field_kwargs.pop('model_field', None)

        if field_class == EnumField:
            return field_class, field_kwargs
        else:
            return super(HyperlinkedModelSerializer, self).build_standard_field(field_name, model_field)





class MeaspkViewSetMixin(object):

    def initial(self, request, *args, **kwargs):
        if kwargs.get('pk') == 'me':
            kwargs['pk'] = request.user.pk

        for k, v in kwargs.items():
            if k.startswith('parent_lookup_') and v == 'me':
                kwargs[k] = request.user.pk

        self.kwargs = kwargs

        super(MeaspkViewSetMixin, self).initial(request, *args, **kwargs)


class ActionSerializerViewSetMixin(object):

    serializer_classes = None

    def get_serializer_class(self):
        if hasattr(self, 'serializer_classes') and isinstance(self.serializer_classes, dict):
            serializer_class = self.serializer_classes.get(self.action)

            if serializer_class:
                return serializer_class

        return super(ActionSerializerViewSetMixin, self).get_serializer_class()

    def get_permissions(self):
        if hasattr(self, 'permission_classes') and isinstance(self.permission_classes, dict):
            permissions = self.permission_classes.get(self.action)

            if permissions:
                return [permission() for permission in permissions]
            else:
                return ()
        else:
            return super(ActionSerializerViewSetMixin, self).get_permissions()
